/*******************************************************************************
 *  lbr.c   |   version 1.22    |   2-Clause BSD License    |   2014-03-11
 *  James Hendrie               |   hendrie dot james at gmail dot com
 *
 *  Description:    By default, this will read in files with long lines and
 *                  break them at 80 characters.
*******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#define LEN_LINE 80
#define LEN_WORD 256
#define LEN_PADDING 0


/*  Optstring
 *      h   help
 *      i   input file
 *      o   output file
 *      w   width (line length)
 *      p   padding
 *      -   read from stdin
 */
static const char *optString = "h-i:o:w:p:";


char *word;     //  Buffer to hold word data

int iLine;      //  Index into line
int iWord;      //  Index into word
int lineLength; //  Length at which to break; default 80
int padding;    //  Padding; default 0
int counter;    //  Generic counter



/*
 * ============================     PRINT HELP      ============================
 */
void print_help( void )
{
    printf("Usage:  lbr [OPTIONS]\n\n");

    printf("This program will, by default, read in text from stdin and break ");
    printf("long lines to\n80 characters or less.\n\n");

    printf("Options:\n");
    printf("  -h or --help\tPrint This help text\n");
    printf("  --version\tPrint version and author info\n");
    printf("  -i\t\tSpecify input file (default stdin)\n");
    printf("  -o\t\tSpecify output file (default stdout)\n");
    printf("  -w\t\tSpecify character width at which to break (default: 80)\n");
    printf("  -p\t\tPad left of line a number of spaces (default:  none)\n\n");

    printf("Examples:\n");
    printf("\tcat TEXT_FILE.txt | lbr\n");
    printf("\t\tPipe text into lbr, print it to stdout (terminal)\n");
    printf("\tlbr -i FILE.txt -o OTHER_FILE.txt -w 76 -p 4\n");
    printf("\t\tRead FILE.txt, break lines at 76 characters, pad the left of ");
    printf("every\n\t\tline with 4 spaces, write to OTHER_FILE.txt\n");
}



/*
 * ============================     PRINT VERSION      =========================
 */
void print_version( void )
{
    printf("lbr version 1.22\n");
    printf("James Hendrie - hendrie.james@gmail.com\n");
}



/*
 * ============================     LBR      ===================================
 */
int lbr( FILE *inFile, FILE *outFile )
{
    char ch;    //  To hold character data received from file

    /*  We test for padding at the very beginning of the output */
    if( padding > 0 )
    {
        ch = fgetc( inFile );
        if( ch != EOF )
        {
            /*  Pad it, baby */
            for( counter = 0; counter < padding; ++counter )
            {
                fputc( ' ', outFile );
                ++iLine;
            }

            /*  Put that first character into our word, otherwise... problems */
            word[ iWord ] = ch;
            ++iWord;
        }

    }


    while( ( ch = fgetc( inFile)) != EOF )
    {
        /*  Put character into word buffer, increment word index */
        word[ iWord ] = ch;
        ++iWord;

        /*
         * If we've hit a space, hit a newline, or the line is getting too long,
         * we need to stop and perform a special action.
         */
        if( ch == ' '  || ch == '\n' || iLine + iWord >= lineLength )
        {
            /*
             * If the line is too long, we need to break it.  Keep in mind that
             * we haven't yet printed the word to the file
             */
            if( iLine + iWord >= lineLength  )
            {
                fputc( '\n', outFile );

                iLine = 0;
                if( padding > 0 )
                {
                    for( counter = 0; counter < padding; ++counter )
                    {
                        fputc( ' ', outFile );
                        ++iLine;
                    }

                }

                fputs( word, outFile );
                iLine += iWord;
                iWord = 0;
                memset( word, '\0', (LEN_WORD) );
            }
            /*  If we hit a newline, just print the word and reset everything */
            else if( ch == '\n' )
            {
                fputs( word, outFile );
                iWord = 0;
                iLine = 0;
                memset( word, '\0', (LEN_WORD) );

                if( padding > 0 )
                {
                    for( counter = 0; counter < padding; ++counter )
                    {
                        fputc( ' ', outFile );
                        ++iLine;
                    }

                }
            }
            /*
             * If the line isn't too long and we haven't hit a newline, we now
             * need to just stick the word in the file and get on with things
             */
            else
            {
                fputs( word, outFile );
                iLine += iWord;
                iWord = 0;
                memset( word, '\0', (LEN_WORD) );
            }
        }
    }

    return(0);
}



/*
 * ============================     CHECK FILE     =============================
 */
int check_file( FILE *fp, const char *filename )
{
    if( fp == NULL )
    {
        fprintf( stderr, "ERROR:  Could not open file %s\n", filename );
        return(1);
    }

    return(0);
}



/*
 * ============================     MAIN     ===================================
 */
int main( int argc, char *argv[] )
{

    /*  Initialize indices */
    iLine = 0;
    iWord = 0;

    /*  Initialize the line length and the padding */
    lineLength = (LEN_LINE);
    padding = (LEN_PADDING);

    /*  Create file pointers */
    FILE *inFile = NULL;
    FILE *outFile = NULL;

    /*  ----------------    ARGUMENT CHECKING   ------------------ */

    /*  Long opts */
    if( argc > 1 )
    {
        if( strcmp( argv[1], "--help") == 0 )
        {
            print_help();
            return(0);
        }
        else if( strcmp( argv[1], "--version") == 0 )
        {
            print_version();
            return(0);
        }
    }



    /*  Short opts */
    int opt = 0;
    opt = getopt( argc, argv, optString );
    while( opt != -1 )
    {
        /*  Check and see if we got a wise-aleck */
        if( ! isalpha(opt) )
        {
            return(1);
        }

        switch( opt )
        {
            case 'h':   //  Print help
                print_help();
                return(0);
                break;
            case 'i':   //  Input file specified
                inFile = fopen( optarg, "r" );
                if( check_file( inFile, optarg ) == 1 )
                    return(1);
                break;
            case '-':   //  Read from stdin
                inFile = stdin;
                break;
            case 'o':   //  Output file specified
                outFile = fopen( optarg, "w+" );
                if( check_file( outFile, optarg ) == 1 )
                    return(1);
                break;
            case 'w':   //  width (line length)
                lineLength = atoi( optarg );
                break;
            case 'p':   //  padding
                padding = atoi( optarg );
                break;
            default:
                /*  Won't be seeing this, disregard */
                break;
        }

        opt = getopt( argc, argv, optString );
    }


    /*  Allocate memory for word buffer, set it all to null char */
    word = malloc( (LEN_WORD) );
    memset( word, '\0', (LEN_WORD) );

    /*  If inFile / outFile haven't been assigned, assign them stdin/stdout */
    if( inFile == NULL )
        inFile = stdin;
    if( outFile == NULL )
        outFile = stdout;

    /*  DO IT !!!! */
    lbr( inFile, outFile );

    /*  Free memory, null pointer */
    free( word );
    word = NULL;

    return(0);
}
